import sys

import argparse

from ucca.ioutil import get_passages_with_progress_bar
from ucca.normalization import normalize
from ucca.validation import validate


def validate_passage(passage, normalization=False, extra=False):
    if normalization:
        normalize(passage, extra=extra)
    return list(validate(passage))


def main(args):
    errors = ((p.ID, validate_passage(p, args.normalize, args.extra))
              for p in get_passages_with_progress_bar(args.filenames, desc="Validating", converters={}))
    errors = {k: v for k, v in errors if v}
    if errors:
        id_len = max(map(len, errors))
        for passage_id, es in sorted(errors.items()):
            for i, e in enumerate(es):
                print("%-*s|%s" % (id_len, "" if i else passage_id, e))
        sys.exit(1)
    else:
        print("No errors found.")


def check_args(parser, args):
    if args.extra and not args.normalize:
        parser.error("Cannot specify --extra without --normalize")
    return args


if __name__ == "__main__":
    argparser = argparse.ArgumentParser(description="Validate UCCA passages")
    argparser.add_argument("filenames", nargs="+", help="files or directories to validate")
    argparser.add_argument("-n", "--normalize", action="store_true", help="normalize before validation")
    argparser.add_argument("-e", "--extra", action="store_true", help="extra normalization rules")
    main(check_args(argparser, argparser.parse_args()))



"""Testing code for the ucca package, unit-testing only."""

import pytest

from ucca import core, layer0, layer1
from .conftest import basic, PASSAGES


def test_creation():
    p = basic()

    assert p.ID == "1"
    assert p.root == p
    assert p.attrib.copy() == {}
    assert p.layer("1").ID == "1"
    assert p.layer("2").ID == "2"
    with pytest.raises(KeyError):
        p.layer("3")

    l1 = p.layer("1")
    l2 = p.layer("2")
    assert l1.root == p
    assert l2.attrib["test"]
    assert l1.orderkey != l2.orderkey
    assert [x.ID for x in l1.all] == ["1.1", "1.2", "1.3"]
    assert [x.ID for x in l1.heads] == ["1.2"]
    assert [x.ID for x in l2.all] == ["2.2", "2.1"]
    assert [x.ID for x in l2.heads] == ["2.2", "2.1"]

    node11, node12, node13 = l1.all
    node22, node21 = l2.all
    assert node11.ID == "1.1"
    assert node11.root == p
    assert node11.layer.ID == "1"
    assert node11.tag == "1"
    assert len(node11) == 0
    assert node11.parents == [node12, node21, node22]
    assert node13.parents == [node12, node22]
    assert node13.attrib.copy() == {"node": True}
    assert len(node12) == 2
    assert node12.children == [node13, node11]
    assert node12[0].attrib.copy() == {"edge": True}
    assert node12.parents == [node22, node21]
    assert node21[0].ID == "2.1->1.1"
    assert node21[1].ID == "2.1->1.2"
    assert node22[0].ID == "2.2->1.1"
    assert node22[1].ID == "2.2->1.2"
    assert node22[2].ID == "2.2->1.3"


def test_modifying():
    p = basic()
    l1, l2 = p.layer("1"), p.layer("2")
    node11, node12, node13 = l1.all
    node22, node21 = l2.all

    # Testing attribute changes
    p.attrib["passage"] = 1
    assert p.attrib.copy() == {"passage": 1}
    del l2.attrib["test"]
    assert l2.attrib.copy() == {}
    node13.attrib[1] = 1
    assert node13.attrib.copy() == {"node": True, 1: 1}
    assert len(node13.attrib) == 2
    assert node13.attrib.get("node")
    assert node13.attrib.get("missing") is None

    # Testing Node changes
    node14 = core.Node(ID="1.4", root=p, tag="4")
    node15 = core.Node(ID="1.5", root=p, tag="5")
    assert l1.all == [node11, node12, node13, node14, node15]
    assert l1.heads == [node12, node14, node15]
    node15.add("test", node11)
    assert node11.parents == [node12, node15, node21, node22]
    node21.remove(node12)
    node21.remove(node21[0])
    assert len(node21) == 0
    assert node12.parents == [node22]
    assert node11.parents == [node12, node15, node22]
    node14.add("test", node15)
    assert l1.heads == [node12, node14]
    node12.destroy()
    assert l1.heads == [node13, node14]
    assert node22.children == [node11, node13]

    node22.tag = "x"
    node22[0].tag = "testx"
    assert node22.tag == "x"
    assert node22[0].tag == "testx"


def test_equals():
    p1 = core.Passage("1")
    p2 = core.Passage("2")
    p1l0 = layer0.Layer0(p1)
    p2l0 = layer0.Layer0(p2)
    p1l1 = layer1.Layer1(p1)
    p2l1 = layer1.Layer1(p2)
    assert (p1.equals(p2) and p2.equals(p1))

    # Checks basic passage equality and Attrib/tag/len differences
    p1l0.add_terminal("0", False)
    p1l0.add_terminal("1", False)
    p1l0.add_terminal("2", False)
    p2l0.add_terminal("0", False)
    p2l0.add_terminal("1", False)
    p2l0.add_terminal("2", False)
    assert (p1.equals(p2) and p2.equals(p1))
    pnct2 = p2l0.add_terminal("3", True)
    assert not (p1.equals(p2) or p2.equals(p1))
    temp = p1l0.add_terminal("3", False)
    assert not (p1.equals(p2) or p2.equals(p1))
    temp.destroy()
    pnct1 = p1l0.add_terminal("3", True)
    assert (p1.equals(p2) and p2.equals(p1))

    # Check Edge and node equality
    ps1 = p1l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    assert not (p1.equals(p2) or p2.equals(p1))
    ps2 = p2l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    assert (p1.equals(p2) and p2.equals(p1))
    p1l1.add_fnode(ps1, layer1.EdgeTags.Participant)
    assert not (p1.equals(p2) or p2.equals(p1))
    assert (ps1.equals(ps2, recursive=False))
    p2l1.add_fnode(ps2, layer1.EdgeTags.Process)
    assert not (p1.equals(p2) or p2.equals(p1))
    p2l1.add_fnode(ps2, layer1.EdgeTags.Participant)
    assert not (p1.equals(p2) or p2.equals(p1))
    p1l1.add_fnode(ps1, layer1.EdgeTags.Process)
    assert (p1.equals(p2) and p2.equals(p1))
    assert not (p1.equals(p2, ordered=True) or
                p2.equals(p1, ordered=True))
    p1l1.add_fnode(ps1, layer1.EdgeTags.Adverbial, implicit=True)
    ps2d3 = p2l1.add_fnode(ps2, layer1.EdgeTags.Adverbial)
    assert not (p1.equals(p2) or p2.equals(p1))
    ps2d3.attrib["implicit"] = True
    assert (p1.equals(p2) and p2.equals(p1))
    ps2[2].attrib["remote"] = True
    assert not (p1.equals(p2) or p2.equals(p1))
    ps1[2].attrib["remote"] = True
    assert (p1.equals(p2) and p2.equals(p1))
    p1l1.add_punct(None, pnct1)
    assert not (p1.equals(p2) or p2.equals(p1))
    p2l1.add_punct(None, pnct2)
    assert (p1.equals(p2) and p2.equals(p1))
    core.Layer("2", p1)
    assert not (p1.equals(p2) or p2.equals(p1))


@pytest.mark.parametrize("create", PASSAGES)
def test_copying(create):
    # we don't need such a complex passage, but it will work anyway
    p1 = create()

    p2 = p1.copy(())
    assert p1.ID == p2.ID
    assert (p1.attrib.equals(p2.attrib))
    assert p1.extra == p2.extra
    assert p1.frozen == p2.frozen

    l0id = layer0.LAYER_ID
    p2 = p1.copy([l0id])
    assert (p1.layer(l0id).equals(p2.layer(l0id)))


def test_iteration():
    p = basic()
    l1, l2 = p.layer("1"), p.layer("2")
    node11, node12, node13 = l1.all
    node22, node21 = l2.all

    assert list(node11.iter()) == [node11]
    assert not list(node11.iter(obj="edges"))
    assert not list(node13.iter(key=lambda x: x.tag != "3"))
    assert list(node12.iter()) == [node12, node13, node11]
    assert list(x.ID for x in node12.iter(obj="edges")) == ["1.2->1.3", "1.2->1.1"]
    assert list(node21.iter(duplicates=True)) == [node21, node11, node12, node13, node11]
    assert list(node21.iter()) == [node21, node11, node12, node13]
    assert list(node22.iter(method="bfs", duplicates=True)) == [node22, node11, node12, node13, node13, node11]



"""Input/output utility functions for UCCA scripts."""
import sys
import time
from collections import defaultdict
from itertools import filterfalse, chain

import os
from glob import glob
from tqdm import tqdm
from xml.etree.ElementTree import ParseError

from ucca.convert import file2passage, passage2file, from_text, to_text, split2segments
from ucca.core import Passage


class LazyLoadedPassages(object):
    """
    Iterable interface to Passage objects that loads files on-the-go and can be iterated more than once
    """
    def __init__(self, files, sentences=False, paragraphs=False, converters=None, lang="en"):
        self.files = files
        self.sentences = sentences
        self.paragraphs = paragraphs
        self.split = self.sentences or self.paragraphs
        self.converters = defaultdict(lambda: from_text) if converters is None else converters
        self.lang = lang
        self._files_iter = None
        self._split_iter = None
        self._file_handle = None
        self._next_index = None

    def __iter__(self):
        self._next_index = 0
        self._files_iter = iter(self.files)
        self._split_iter = None
        self._file_handle = None
        return self

    def __next__(self):
        passage = self._next_passage()
        self._next_index += 1
        return passage

    def _next_passage(self):
        passage = None
        if self._split_iter is None:
            try:
                file = next(self._files_iter)
            except StopIteration:  # Finished iteration
                raise
            if isinstance(file, Passage):  # Not really a file, but a Passage
                passage = file
            else:  # A file
                attempts = 3
                while not os.path.exists(file):
                    with tqdm.external_write_mode(file=sys.stderr):
                        if attempts == 0:
                            print("File not found: %s" % file, file=sys.stderr)
                            return next(self)
                        print("Failed reading %s, trying %d more times..." % (file, attempts), file=sys.stderr)
                    time.sleep(5)
                    attempts -= 1
                try:
                    passage = file2passage(file)  # XML or binary format
                except (IOError, ParseError):  # Failed to read as passage file
                    base, ext = os.path.splitext(os.path.basename(file))
                    converter = self.converters.get(ext.lstrip("."))
                    if converter is None:
                        raise
                    self._file_handle = open(file, encoding="utf-8")
                    self._split_iter = iter(converter(chain(self._file_handle, [""]), passage_id=base, lang=self.lang))
            if self.split:
                if self._split_iter is None:
                    self._split_iter = (passage,)
                self._split_iter = iter(s for p in self._split_iter for s in
                                        split2segments(p, is_sentences=self.sentences, lang=self.lang))
        if self._split_iter is not None:  # Either set before or initialized now
            try:
                passage = next(self._split_iter)
            except StopIteration:  # Finished this converter
                self._split_iter = None
                if self._file_handle is not None:
                    self._file_handle.close()
                    self._file_handle = None
                return next(self)
        return passage

    # The following three methods are implemented to support shuffle;
    # note files are shuffled but there is no shuffling within files, as it would not be efficient.
    # Note also the inconsistency because these access the files while __iter__ accesses individual passages.
    def __len__(self):
        return len(self.files)

    def __getitem__(self, i):
        return self.files[i]

    def __setitem__(self, i, value):
        self.files[i] = value

    def __bool__(self):
        return bool(self.files)


def get_passages_with_progress_bar(filename_patterns, desc=None, converters=None):
    t = tqdm(get_passages(filename_patterns, converters=converters), desc=desc, unit=" passages")
    for passage in t:
        t.set_postfix(ID=passage.ID)
        yield passage


def get_passages(filename_patterns, converters=None):
    for pattern in [filename_patterns] if isinstance(filename_patterns, str) else filename_patterns:
        for filenames in glob(pattern) or [pattern]:
            yield from read_files_and_dirs(filenames, converters=converters)


def gen_files(files_and_dirs):
    """
    :param files_and_dirs: iterable of files and/or directories to look in
    :return: all files given, plus any files directly under any directory given
    """
    for file_or_dir in [files_and_dirs] if isinstance(files_and_dirs, str) else files_and_dirs:
        if os.path.isdir(file_or_dir):
            yield from filterfalse(os.path.isdir, (os.path.join(file_or_dir, f) for f in os.listdir(file_or_dir)))
        else:
            yield file_or_dir


def read_files_and_dirs(files_and_dirs, sentences=False, paragraphs=False, converters=None, lang="en"):
    """
    :param files_and_dirs: iterable of files and/or directories to look in
    :param sentences: whether to split to sentences
    :param paragraphs: whether to split to paragraphs
    :param converters: dict of input format converters to use based on the file extension
    :param lang: language to use for tokenization model
    :return: lazy-loaded passages from all files given, plus any files directly under any directory given
    """
    return LazyLoadedPassages(list(gen_files(files_and_dirs)), sentences=sentences, paragraphs=paragraphs,
                              converters=converters, lang=lang)


def write_passage(passage, output_format=None, binary=False, outdir=".", prefix="", converter=None, verbose=True,
                  append=False):
    """
    Write a given UCCA passage in any format.
    :param passage: Passage object to write
    :param output_format: filename suffix (if given "ucca", suffix will be ".pickle" or ".xml" depending on `binary')
    :param binary: save in pickle format with ".pickle" suffix
    :param outdir: output directory, should exist already
    :param prefix: string to prepend to output filename
    :param converter: function to apply to passage before saving (if output_format is not "ucca"/"pickle"/"xml"),
                      returning iterable of strings, each corresponding to an output line
    :param verbose: print "Writing passage" message
    :param append: if using converter, append to output file rather than creating a new file
    :return: path of created output file
    """
    suffix = output_format if output_format and output_format != "ucca" else ("pickle" if binary else "xml")
    outfile = os.path.join(outdir, prefix + passage.ID + "." + suffix)
    if verbose:
        with tqdm.external_write_mode():
            print("Writing passage '%s'..." % outfile)
    if output_format is None or output_format in ("ucca", "pickle", "xml"):
        passage2file(passage, outfile, binary=binary)
    else:
        with open(outfile, "a" if append else "w", encoding="utf-8") as f:
            f.writelines(map("{}\n".format, (converter or to_text)(passage)))
    return outfile



#!/usr/bin/env python3

import argparse
import sys

from tqdm import tqdm

from ucca import convert
from ucca.evaluation import evaluate, Scores
from ucca.ioutil import get_passages_with_progress_bar

desc = """Parses files in CoNLL-X, SemEval 2015 SDP, NeGra export or text format,
converts to UCCA standard format, converts back to the original format and evaluates."""


def main(args):
    converter1 = convert.TO_FORMAT[args.format]
    converter2 = convert.FROM_FORMAT[args.format]
    scores = []
    for ref in get_passages_with_progress_bar(args.filenames, desc="Converting"):
        try:
            guessed = next(converter2(converter1(ref, tree=args.tree), ref.ID))
            scores.append(evaluate(guessed, ref, verbose=args.verbose))
        except Exception as e:
            if args.strict:
                raise ValueError("Error evaluating conversion of %s" % ref.ID) from e
            else:
                with tqdm.external_write_mode():
                    print("Error evaluating conversion of %s: %s" % (ref.ID, e), file=sys.stderr)
    print()
    if args.verbose and len(scores) > 1:
        print("Aggregated scores:")
    Scores.aggregate(scores).print()


if __name__ == '__main__':
    argparser = argparse.ArgumentParser(description=desc)
    argparser.add_argument("filenames", nargs="+", help="file names to convert and evaluate")
    argparser.add_argument("-f", "--format", required=True, choices=convert.CONVERTERS, help="input file format")
    argparser.add_argument("-T", "--tree", action="store_true", help="remove multiple parents to get a tree")
    argparser.add_argument("-s", "--strict", action="store_true",
                           help="stop immediately if failed to convert or evaluate a file")
    argparser.add_argument("-v", "--verbose", action="store_true",
                           help="print evaluation results for each file separately")
    main(argparser.parse_args())
